#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <omp.h>

#include "io_vanilla.h"
#include "generate_vanilla.h"

/* avoid *[]-confusion */
typedef double double3[3];

/*
 * Global constants
 */

/* set gravity constant to one */
const double G = 1.0;
/* The epsilon^2 for the softening */
const double epsilon2 = 1E-14;
/* Make the tile size a constant to allow better loop-unrolling by the compiler */
const int B = 1;

#define MIN(x,y) (((x)<(y)) ? (x) : (y))
#define MAX(x,y) (((x)>(y)) ? (x) : (y))

/**
 * Compute acceleration
 *
 * \param[in]  n Number of bodies
 * \param[in]  x Positions of the bodies
 * \param[in]  m Masses of the bodies
 * \param[out] a The computed accelerations of the bodies
 *
 * The inner loop executes 26 floating point operations (for simplicity
 * we count sqrt and / as 1 flop each). Then, each inner loop is
 * executed n * (n-1) / 2 times (geometric sum formula summing over the number
 * of upper-right non-diagonal entries of the 2D-field traversed by
 * indices i and j), resulting in 13 * (n+1) * n floating point operations
 * total * per call to acceleration().
 */
void acceleration_openmp(int n, double3 x[], double m[], double3 a[])
{
  int I,J,i,j,iend,jstart,jend;
  double d0,d1,d2,r,r2,factori,factorj;
  double invfact;
  double X[B][6], A[B][6], M[B][2];

  /* compute acceleration exploiting symmetry */
  #pragma omp parallel for private(J,i,j,iend,jstart,jend,d0,d1,d2,r,r2,factori,factorj,invfact,X,A,M) 
  for (I=0; I<n; I+=B)
    for (J=I; J<n; J+=B) {
	  /* load to avoid aliasing */
	  iend = MIN(I+B,n);
	  jend = MIN(J+B,n);
	  for (i=I; i<iend; i++) {
        X[i-I][0] = x[i][0]; X[i-I][1] = x[i][1]; X[i-I][2] = x[i][2];
        A[i-I][0] = 0.0;     A[i-I][1] = 0.0;     A[i-I][2] = 0.0;
        M[i-I][0] = m[i];
	  }
	  for (j=J; j<jend; j++) {
        X[j-J][3] = x[j][0]; X[j-J][4] = x[j][1]; X[j-J][5] = x[j][2];
        A[j-J][3] = 0.0;     A[j-J][4] = 0.0;     A[j-J][5] = 0.0;
        M[j-J][1] = m[j];
	  }

      /* compute locally */
	  for (i=0; i<iend-I; i++) {
        jstart = MAX(J,I+i+1)-J; 
	    for (j=jstart; j<jend-J; j++) {
          /* compute distance vector */
          d0 = X[j][3]-X[i][0];
          d1 = X[j][4]-X[i][1];
          d2 = X[j][5]-X[i][2];

          /* compute (scalar valued) distance and distance^2, with softening */
          r2 = d0*d0 + d1*d1 + d2*d2 + epsilon2;
		  r = sqrt(r2);

          /* precompute the factors in the acceleration equation */
          invfact = G/(r*r2);
		  factori = M[i][0]*invfact;
		  factorj = M[j][1]*invfact;
		  
          /* compute acceleration of body i due to body j */
          A[i][0] += factorj*d0;
          A[i][1] += factorj*d1;
          A[i][2] += factorj*d2;

          /* compute acceleration of body j due to body i */
          A[j][3] -= factori*d0;
          A[j][4] -= factori*d1;
          A[j][5] -= factori*d2;
		}
	  }

      #pragma omp critical
      {
        /* store result */
        for (i=I; i<iend; i++) {
	      a[i][0] += A[i-I][0]; a[i][1] += A[i-I][1]; a[i][2] += A[i-I][2];
	    }
        for (j=J; j<jend; j++) {
	      a[j][0] += A[j-J][3]; a[j][1] += A[j-J][4]; a[j][2] += A[j-J][5];
        }
	  }
	}
}

/**
 * Do one time step
 *
 * \param[in]     n    Number of bodies
 * \param[in]     dt   Time step size
 * \param[in,out] x    IN: old positions, OUT: new positions
 * \param[in,out] v    IN: old velocities, OUT: new velocities
 * \param[in]     m    Masses
 * \param[in,out] a    IN: old accelerations, OUT: new accelerations
 * \param         aold Buffer to hold the old accelerations.  Input and output
 *                     contents is not meaningful.
 *
 * The "update position" and "update velocity" execute 12 floating point
 * operations per iteration each, resulting in 24*n floating point ops
 * together.  Add to it the 3 flops to precompute time step related values and
 * the 13*n*(n-1) flops from acceleration() to arrive at the total of
 * 13*n*(n-1)+24*n+3 floating point operations per call to leapfrog().
 */
void leapfrog (int n, double dt, double3 x[], double3 v[], double m[],
               double3 a[], double3 aold[])
{
  int i;
  /* precompute some values from the time step */
  double dt2 = dt*dt*0.5;
  double dthalf = dt*0.5;

  /* update position */
  for (i=0; i<n; i++) {
    x[i][0] += dt*v[i][0] + dt2*a[i][0];
    x[i][1] += dt*v[i][1] + dt2*a[i][1];
    x[i][2] += dt*v[i][2] + dt2*a[i][2];
  }

  /* save and clear acceleration */
  for (i=0; i<n; i++) {
    aold[i][0] = a[i][0];
    aold[i][1] = a[i][1];
    aold[i][2] = a[i][2];
    a[i][0] = a[i][1] = a[i][2] = 0.0;
  }

  /* compute new acceleration */
  acceleration_openmp(n,x,m,a);

  /* update velocity */
  for (i=0; i<n; i++) {
    v[i][0] += dthalf*aold[i][0] + dthalf*a[i][0];
    v[i][1] += dthalf*aold[i][1] + dthalf*a[i][1];
    v[i][2] += dthalf*aold[i][2] + dthalf*a[i][2];
  }
}

/**
 * main program
 */
int main (int argc, char** argv)
{
  /* hold positions, velocities, masses, and accelerations for the bodies */
  double3 *x;
  double3 *v;
  double *m;
  double3 *a;
  /* temporary buffer given to leapfrog() so it can store the old acceleration
     values there */
  double3 *temp;
  /* loop counter */
  int i;
  /* count the time steps done */
  int k = 0;
  /* base name for the output files */
  char base[]="openmp";
  /* buffer to construct the whole output filename in */
  char name[256];
  /* handle for the output file */
  FILE *file;
  /* accumulated simulation time and time step size.  These will be modified
     when reading initial values from a file */
  double t = 0, dt = 1E-3;
  /* for measuring the elapsed time and calculating the MFLOP rate */
  double start,stop,elapsed,flop;
  /* number of bodies (read from commandline) */
  int n;
  /* number of time steps to do (read from commandline) */
  int timesteps;
  /* when to measure MFLOP rate and write output files: every mod time steps
     (read from commandline) */
  int mod;
  /* write vtk files ? */
  int output = 0;
  /* number of OMP-threads */
  int numthreads;

  /**
   * parse commandline arguments
   */
  if (argc!=5) {
    printf("usage: nbody_openmp <nbodies> <timesteps> <every> <numthreads>\n");
    return 1;
  }
  sscanf(argv[1],"%d",&n);
  sscanf(argv[2],"%d",&timesteps);
  sscanf(argv[3],"%d",&mod);
  sscanf(argv[4],"%d",&numthreads);

#ifdef _OPENMP
  printf("number of processors: %d\n",omp_get_num_procs());
  omp_set_num_threads(numthreads);
  printf("number of threads used: %d\n",omp_get_max_threads());
#else
  printf("not compiled with OpenMP, set numtrheads to 1!\n");
  omp_set_num_threads(1);
#endif

#if 0
  /* get initial value from a data file */
  file = fopen("plummer10000.vtk","r");
  if (file==NULL) {
    printf("could not open file --- aborting\n");
    return 1;
  }
  /* peek into the file to find out the number of bodies */
  n = get_vtk_numbodies(file);
  rewind(file);
  /* allocate memory accordingly */
  x = calloc(n,sizeof(double3));
  v = calloc(n,sizeof(double3));
  m = calloc(n,sizeof(double));
  /* read everything */
  read_vtk_file_double(file,n,x,v,m,&t,&dt);
  fclose(file);
  printf("loaded %d bodies\n",n);
#else
  /* get initial values from one of the generator functions */
  x = calloc(n,sizeof(double3));
  v = calloc(n,sizeof(double3));
  m = calloc(n,sizeof(double));
  plummer(n,17,x,v,m);
  /* cube(n,17,1000,1.0,0.2,x,v,m); */
#endif

  /**
   * allocate and initialise the acceleration and the buffer for leapfrog
   */
  a = calloc(n,sizeof(double3));
  for (i=0; i<n; i++)
    a[i][0] = a[i][1] = a[i][2] = 0.0;
  acceleration_openmp(n,x,m,a);
  temp = calloc(n,sizeof(double3));

  /* write out the initial values */
  if (output == 0) {
    sprintf(name,"%s_%06d.vtk",base,k/mod);
    printf("writing %s \n",name);
    file = fopen(name,"w");
    write_vtk_file_double(file,n,x,v,m,t,dt);
    fclose(file);
}

  /* start the time measurement */
  start = omp_get_wtime();
  for (k=1; k<timesteps; k++) {
    /* do one timestep and increment elapsed simulated time */
    leapfrog(n,dt,x,v,m,a,temp);
    t += dt;
    if (k%mod==0) {
      /* write statistics */
      stop = omp_get_wtime();
      elapsed = stop-start;
      /* 13 * n * (n-1) + 24*n + 3 FLOP from leapfrog(), 1 from the t+=dt above */
      flop = mod * (13.0 * n * (n - 1.0) + 24.0 * n + 4.0);
      printf("%g seconds for %g ops = %g MFLOPS\n", elapsed, flop, flop/elapsed/1E6);

      /* write output file */
      if (output == 0) {
        sprintf(name,"%s_%06d.vtk",base,k/mod);
        printf("writing %s \n",name);
        file = fopen(name,"w");
        write_vtk_file_double(file,n,x,v,m,t,dt);
        fclose(file);
      }

      /* restart the timer */
      start = omp_get_wtime();
    }
  }

  /* Don't bother freeing the allocated buffers since the program is going to
     end and the operating system will reclaim all of its memory anyway */

  return 0;
}
